# -*- coding: utf-8 -*-
"""
    
    This file is part of pyEnsemblRest.
    Copyright (C) 2013-2016, Steve Moss

    pyEnsemblRest is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    pyEnsemblRest is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with pyEnsemblRest.  If not, see <http://www.gnu.org/licenses/>.

    This script allows testing of all pyEnsemblRest functionality.

"""

# import my module
import ensemblrest

# import other modules
import six
import re
import json
import time
import shlex
import urllib
import logging
import subprocess
import unittest

#logger instance
logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)

# create console handler and set level to debug. NullHandler to put all into /dev/null
#ch = logging.NullHandler()

# This console handle write all logging to and opened strem. sys.stderr is the default
ch = logging.StreamHandler()

# Set the level for this handler
ch.setLevel(logging.DEBUG)

# create formatter
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')

# add formatter to ch
ch.setFormatter(formatter)

# add ch to logger
logger.addHandler(ch)

# Wait some time before next request
WAIT = 0.5

# Sometimes curl fails
MAX_RETRIES = 5

# curl timeouts
TIMEOUT = 60


def launch(cmd):
    """calling a cmd with subprocess"""
    
    # setting curl timeouts
    pattern = re.compile("curl")
    repl = "curl --connect-timeout %s --max-time %s" %(TIMEOUT, TIMEOUT*2)
    
    # Setting curl options
    cmd = re.sub(pattern, repl, cmd)
    
    logger.debug("Executing: %s" %(cmd))
    
    args = shlex.split(cmd)
    p = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)
    stdout, stderr = p.communicate()
    
    if len(stderr) > 0: 
        logger.debug(stderr)
        
    # debug
    logger.debug("Got: %s" %(stdout))
        
    return stdout

def jsonFromCurl(curl_cmd):
    """Parsing a JSON curl result"""

    data = None
    retry = 0
    
    while retry < MAX_RETRIES:
        # update retry
        retry += 1
        
        # execute the curl cmd
        result = launch(curl_cmd)
        
        # load it as a dictionary
        try:
            data = json.loads(result)
            
        except ValueError as e:
            logger.warning("Curl command failed: %s" % e)
            time.sleep(WAIT*10)
            
            #next request
            continue

        if isinstance(data, dict):
            if "error" in data:
                logger.warning("Curl command failed: %s" %(data["error"]))
                time.sleep(WAIT*10)
                
                #next request
                continue
                
        # If I arrive here, I assume that curl went well
        break
    
    return data

def _genericCMP(v1, v2):
    """Check ensembl complex elements"""
    
    logger.debug("Check %s == %s" %(v1, v2))
    
    # check that types are the same
    if type(v1) != type(v2):
        return False
        
    elif isinstance(v1, dict):
        #call comparedict
        if compareDict(v1, v2) is False:
            return False
                    
    elif isinstance(v1, list):
        #call comparedict
        if compareList(v1, v2) is False:
            return False
            
    elif isinstance(v1, (six.string_types, float, int)):
        if v1 != v2:
            return False
        
    else:
        logger.error("%s <> %s" %(v1, v2))
        logger.critical("Case not implemented: type:%s" %(type(v1)))
    
    #default value
    return True

# A function to evaluate if two python complex dictionaries are the same
def compareDict(d1, d2):
    """A function to evaluate if two python complex dictionaries are the same"""
    if d1 == d2:
        return True
        
    # check keys
    k1 = d1.keys()
    k2 = d2.keys()
    
    # sorting keys
    k1 = sorted(k1)
    k2 = sorted(k2)

    logger.debug(k1)
    logger.debug(k2)
    
    # check keys are equals
    if k1 != k2:
        return False
        
    #now I have to check values for each key value
    for k in k1:
        #get values
        v1 = d1[k]
        v2 = d2[k]
        
        if v1 == v2:
            continue
        
        # the species key may differ in some cases: ex: Tgut-Mgal-Ggal[3] <> Ggal-Mgal-Tgut[3]
        if k in ["species", "tree"] and isinstance(v1, six.string_types) and isinstance(v2, six.string_types):
            pattern = re.compile("([\w]+)-?(?:\[\d\])?")
            
            #override values
            v1 = re.findall(pattern, v1)
            v2 = re.findall(pattern, v2)
            
        # check if elements are the same
        if _genericCMP(v1, v2) is False:
            return False
            
    #if I arrive here:
    return True

def compareList(l1, l2):
    """A function to evaluate if two python complex list are the same"""
    
    if l1 == l2:
        return True
        
    #check lengths
    if len(l1) != len(l2):
        return False
        
    #I cannot use set nor collections.Count, since elements could't be hashable
    # sorting elements doesn't apply since elements may be un-hashable
    for i in range(len(l1)):
        v1 = l1[i]
        
        flag_found = False
        
        for j in range(len(l2)):
            v2 = l2[j]
        
            if v1 == v2:
                flag_found = True
            
            # check if elements are the same
            elif _genericCMP(v1, v2) is True:
                flag_found = True
                
            #If I found en equal element, i can stop
            if flag_found is True:
                break
            
        #After cycling amoung l2, if I can't find an equal element
        if flag_found is False:
            return False
        
    #if I arrive here
    return True


class EnsemblRest(unittest.TestCase):
    """A class to test EnsemblRest methods"""
    
    def setUp(self):
        """Create a EnsemblRest object"""
        self.EnsEMBL = ensemblrest.EnsemblRest()
        
    def tearDown(self):
        """Sleep a while before doing next request"""
        time.sleep(WAIT)

class EnsemblRestBase(EnsemblRest):
    """A class to deal with ensemblrest base methods"""
        
    def test_setHeaders(self):
        """Testing EnsemblRest with no headers provided"""
        
        user_agent = ensemblrest.ensembl_config.ensembl_user_agent
        self.EnsEMBL = ensemblrest.EnsemblRest(headers={})
        self.assertEqual(self.EnsEMBL.session.headers.get("User-Agent"), user_agent)
        
    def test_mandatoryParameters(self):
        """Testing EnsemblRest with no mandatory parameters"""
        
        six.assertRaisesRegex(self, Exception, "mandatory param .* not specified", self.EnsEMBL.getArchiveById)
        
    def test_wait4request(self):
        """Simulating max request per second"""
        
        self.EnsEMBL.getArchiveById(id='ENSG00000157764')
        self.EnsEMBL.req_count = 15
        self.EnsEMBL.last_req += 2
        self.EnsEMBL.getArchiveById(id='ENSG00000157764')
        
    def test_methodNotImplemented(self):
        """Testing a not implemented method"""
        
        #Add a non supported method
        ensemblrest.ensemblrest.ensembl_api_table["notImplemented"] = {
            'doc' : 'Uses the given identifier to return the archived sequence',
            'url': '/archive/id/{{id}}',
            'method': 'HEAD',
            'content_type': 'application/json'
        }
        
        # register this method
        self.EnsEMBL.__dict__["notImplemented"] = self.EnsEMBL.register_api_func("notImplemented", ensemblrest.ensembl_config.ensembl_api_table)
            
        #Set __doc__ for generic class method
        self.EnsEMBL.__dict__["notImplemented"].__doc__ = ensemblrest.ensemblrest.ensembl_api_table["notImplemented"]["doc"]
        
        #add function name to the class methods
        self.EnsEMBL.__dict__["notImplemented"].__name__ = "notImplemented"
        
        # call the new function and deal with the exception
        self.assertRaises(NotImplementedError, self.EnsEMBL.notImplemented, id='ENSG00000157764') 
        
    def __something_bad(self, curl_cmd, last_response):
        """A function to test 'something bad' message"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
        
        # create a fake request.Response class
        class FakeResponse():
            def __init__(self, response):
                self.headers = response.headers
                self.status_code = 400
                self.text = """{"error":"something bad has happened"}"""
                
        #instantiate a fake response
        fakeResponse = FakeResponse(last_response)
        test = self.EnsEMBL.parseResponse(fakeResponse)
        
        # testing values
        self.assertDictEqual(reference, test)
        self.assertGreaterEqual(self.EnsEMBL.last_attempt, 1)
    
    def test_SomethingBad(self):
        """Deal with the {"error":"something bad has happened"} message"""
        
        # get the curl cmd from ensembl site:
        curl_cmd = "curl 'http://rest.ensembl.org/archive/id/ENSG00000157764?' -H 'Content-type:application/json'"
        
        # get a request
        self.EnsEMBL.getArchiveById(id="ENSG00000157764")
        
        # retrieve last_reponse
        last_response = self.EnsEMBL.last_response
        
        # call generic function
        self.__something_bad(curl_cmd, last_response)
        
    def test_SomethingBadPOST(self):
        """Deal with the {"error":"something bad has happened"} message using a POST method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/lookup/id' -H 'Content-type:application/json' -H 'Accept:application/json' -X POST -d '{ "ids" : ["ENSG00000157764", "ENSG00000248378" ] }'"""
      
        # execute EnsemblRest function
        self.EnsEMBL.getLookupByMultipleIds(ids=["ENSG00000157764", "ENSG00000248378" ])
        
        # retrieve last_reponse
        last_response = self.EnsEMBL.last_response
        
        # call generic function
        self.__something_bad(curl_cmd, last_response)
        
    def test_LDFeatureContainerAdaptor(self):
        """Deal with the {"error":"Something went wrong while fetching from LDFeatureContainerAdaptor"} message"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/ld/human/pairwise/rs6792369/rs1042779?population_name=1000GENOMES:phase_3:KHV;r2=0.85' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
        
        # get a request
        self.EnsEMBL.getLdPairwise(species="human", id1="rs6792369", id2="rs1042779", population_name="1000GENOMES:phase_3:KHV", r2=0.85)
        
        # retrieve last_reponse
        response = self.EnsEMBL.last_response
        
        # create a fake request.Response class
        class FakeResponse():
            def __init__(self, resp):
                self.headers = resp.headers
                self.status_code = 400
                self.text = """{"error":"Something went wrong while fetching from LDFeatureContainerAdaptor"}"""
                
        #instantiate a fake response
        fakeResponse = FakeResponse(response)
        test = self.EnsEMBL.parseResponse(fakeResponse)
        
        # testing values
        self.assertEqual(reference, test)
        self.assertGreaterEqual(self.EnsEMBL.last_attempt, 1)
        

class EnsemblRestArchive(EnsemblRest):
    """A class to deal with ensemblrest archive methods"""
    
    def test_getArchiveById(self):
        """Test archive GET endpoint"""
      
        # get the curl cmd from ensembl site:
        curl_cmd = "curl 'http://rest.ensembl.org/archive/id/ENSG00000157764?' -H 'Content-type:application/json'"
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getArchiveById(id='ENSG00000157764')
        
        # testing values
        self.assertDictEqual(reference, test)
        
    def test_getXMLArchiveById(self):
        """text archive GET endpoint returning XML"""
        
        # get the curl cmd from ensembl site:
        curl_cmd = "curl 'http://rest.ensembl.org/archive/id/ENSG00000157764?' -H 'Content-type:text/xml'"
        
        # execute the curl cmd an get data as a dictionary
        reference = launch(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getArchiveById(id='ENSG00000157764', content_type="text/xml")
        
        # testing values
        self.assertEqual(reference, test)
        
    def test_getArchiveByMultipleIds(self):
        """Test archive POST endpoint"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/archive/id' -H 'Content-type:application/json' \
-H 'Accept:application/json' -X POST -d '{ "id" : ["ENSG00000157764", "ENSG00000248378"] }'"""

        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getArchiveByMultipleIds(id=["ENSG00000157764", "ENSG00000248378"])
        
        # testing values
        self.assertListEqual(reference, test)
    
class EnsemblRestComparative(EnsemblRest):
    """A class to deal with ensemblrest comparative genomics methods"""
    
    def test_getGeneTreeById(self):
        """Test genetree by id GET method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/genetree/id/ENSGT00390000003602?' -H 'Content-type:application/json'"""
                
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function. Dealing with application/json is simpler, since text/x-phyloxml+xml may change elements order
        test = self.EnsEMBL.getGeneTreeById(id='ENSGT00390000003602', content_type="application/json")
        
        # testing values
        self.assertTrue(compareDict(reference, test))
        
    def test_getGeneTreeMemberById(self):
        """Test genetree by member id GET method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/genetree/member/id/ENSG00000157764?prune_species=cow;prune_taxon=9526' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function. Dealing with application/json is simpler, since text/x-phyloxml+xml may change elements order
        test = self.EnsEMBL.getGeneTreeMemberById(id='ENSG00000157764', prune_species="cow", prune_taxon=9526, content_type="application/json")

        # Set self.maxDiff to None to see differences
        # self.maxDiff = None
        
        # testing values. Since json are nested dictionary and lists, and they are not hashable, I need to order list before
        # checking equality, and I need to ensure that dictionaries have the same keys and values
        self.assertTrue(compareDict(reference, test))
        
    def test_getGeneTreeMemberBySymbol(self):
        """Test genetree by symbol GET method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/genetree/member/symbol/homo_sapiens/BRCA2?prune_species=cow;prune_taxon=9526' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function. Dealing with application/json is simpler, since text/x-phyloxml+xml may change elements order
        test = self.EnsEMBL.getGeneTreeMemberBySymbol(species='human', symbol='BRCA2', prune_species="cow", prune_taxon=9526, content_type="application/json")
        
        # testing values
        self.assertDictEqual(reference, test)
        
    def test_getAlignmentByRegion(self):
        """Test get genomic alignment region GET method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/alignment/region/taeniopygia_guttata/2:106040000-106040050:1?species_set_group=sauropsids' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function. Dealing with application/json is simpler
        test = self.EnsEMBL.getAlignmentByRegion(species="taeniopygia_guttata", region="2:106040000-106040050:1", species_set_group="sauropsids")
        
        # testing values. Values in list can have different order
        self.assertTrue(compareList(reference, test))
        
    def test_getHomologyById(self):
        """test get homology by Id GET method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/homology/id/ENSG00000157764?' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function. Dealing with application/json is simpler, since text/x-phyloxml+xml may change elements order
        test = self.EnsEMBL.getHomologyById(id='ENSG00000157764')
        
        # testing values. Since json are nested dictionary and lists, and they are not hashable, I need to order list before
        # checking equality, and I need to ensure that dictionaries have the same keys and values
        self.assertTrue(compareDict(reference, test))
        
    def test_getHomologyBySymbol(self):
        """test get homology by symbol"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/homology/symbol/human/BRCA2?' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function. Dealing with application/json is simpler, since text/x-phyloxml+xml may change elements order
        test = self.EnsEMBL.getHomologyBySymbol(species='human', symbol='BRCA2')
        
        # testing values. Since json are nested dictionary and lists, and they are not hashable, I need to order list before
        # checking equality, and I need to ensure that dictionaries have the same keys and values
        self.assertTrue(compareDict(reference, test))
    
class EnsemblRestXref(EnsemblRest):
    """A class to deal with ensemblrest cross references methods"""
    
    def test_getXrefsBySymbol(self):
        """Testing get XRef by Id GET method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/xrefs/symbol/homo_sapiens/BRCA2?' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getXrefsBySymbol(species='human', symbol='BRCA2')
        
        # testing values
        self.assertEqual(reference, test)
        
    def test_getXrefsByName(self):
        """Testing get XRef by Id GET method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/xrefs/name/human/BRCA2?' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getXrefsByName(species='human', name='BRCA2')
        
        # testing values
        self.assertEqual(reference, test)
        
    def test_getXrefsById(self):
        """Testing get XRef by Id GET method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/xrefs/id/ENSG00000157764?' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getXrefsById(id='ENSG00000157764')
        
        # testing values
        self.assertEqual(reference, test)
        
    
class EnsemblRestInfo(EnsemblRest):
    """A class to deal with ensemblrest information methods"""

    def test_getInfoAnalysis(self):
        """Testing Info analysis GET method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/info/analysis/homo_sapiens?' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getInfoAnalysis(species="homo_sapiens")
        
        # testing values
        self.assertEqual(reference, test)
    
    def test_getInfoAssembly(self):
        """Testing Info assembly GET method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/info/assembly/homo_sapiens?' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getInfoAssembly(species="homo_sapiens")
        
        # testing values
        self.assertEqual(reference, test)
        
    def test_getInfoAssemblyRegion(self):
        """Testing Info Assembly by region GET method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/info/assembly/homo_sapiens/X?' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getInfoAssemblyRegion(species="homo_sapiens", region_name="X")
        
        # testing values
        self.assertEqual(reference, test)
        
    def test_getInfoBiotypes(self):
        """Testing Info BioTypes GET method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/info/biotypes/homo_sapiens?' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getInfoBiotypes(species="homo_sapiens")
        
        # testing values
        self.assertEqual(reference, test)
        
    def test_getInfoComparaMethods(self):
        """Testing Info Compara Methods GET method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/info/compara/methods/?' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getInfoComparaMethods()
        
        # testing values
        self.assertEqual(reference, test)
        
    def test_getInfoComparaSpeciesSets(self):
        """Testing Info Compara Species Sets GET method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/info/compara/species_sets/EPO?' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getInfoComparaSpeciesSets(methods="EPO")
        
        # testing values
        self.assertEqual(reference, test)
        
    def test_getInfoComparas(self):
        """Testing Info Compara GET method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/info/comparas?' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getInfoComparas()
        
        # testing values
        self.assertEqual(reference, test)
        
    def test_getInfoData(self):
        """Testing Info Data GET method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/info/data/?' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getInfoData()
        
        # testing values
        self.assertEqual(reference, test)
    
    def test_getInfoExternalDbs(self):
        """Testing Info External Dbs GET method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/info/external_dbs/homo_sapiens?' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getInfoExternalDbs(species="homo_sapiens")
        
        # testing values
        self.assertEqual(reference, test)

    def test_getInfoPing(self):
        """Testing Info Ping GET method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/info/ping?' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getInfoPing()
        
        # testing values
        self.assertEqual(reference, test)
    
    def test_getInfoRest(self):
        """Testing Info REST GET method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/info/rest?' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getInfoRest()
        
        # testing values
        self.assertEqual(reference, test)
    
    def test_getInfoSoftware(self):
        """Testing Info Software GET method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/info/software?' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getInfoSoftware()
        
        # testing values
        self.assertEqual(reference, test)
    
    def test_getInfoSpecies(self):
        """Testing Info Species GET method"""        
        
        curl_cmd = """curl 'http://rest.ensembl.org/info/species?division=ensembl' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getInfoSpecies(division="ensembl")
        
        try:
            # testing values. Since json are nested dictionary and lists, and they are not hashable, I need to order list before
            # checking equality, and I need to ensure that dictionaries have the same keys and values
            self.assertTrue(compareDict(reference, test))
        
        # The transitory failure seems to be related to a misconfiguration of ensembl
        # rest service. In such cases is better to inform dev<at>ensembl.org and report
        # such issues
        except AssertionError as e:
            # sometimes this test can fail. In such case, i log the error
            logger.error(e)
            logger.error(
                "Sometimes 'test_getInfoSpecies' fails. This could be a transitory problem on EnsEMBL REST service"
            )
        
    def test_getInfoVariation(self):
        """Testing Info Variation GET method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/info/variation/homo_sapiens?' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getInfoVariation(species="homo_sapiens")
        
        # testing values
        self.assertEqual(reference, test)
        
    def test_getInfoVariationPopulations(self):
        """Testing Info Variation Populations GET method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/info/variation/populations/homo_sapiens?filter=LD' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getInfoVariationPopulations(species="homo_sapiens", filter="LD")
        
        # testing values
        self.assertEqual(reference, test)
        
    
class EnsemblRestLinkage(EnsemblRest):
    """A class to deal with ensemblrest linkage disequilibrium methods"""

    def test_getLdId(self):
        """Testing get LD ID GET method"""

        curl_cmd = """curl 'http://rest.ensembl.org/ld/human/rs1042779/1000GENOMES:phase_3:KHV?window_size=10;d_prime=1.0' -H 'Content-type:application/json'"""

        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getLdId(species="human", id="rs1042779", population_name="1000GENOMES:phase_3:KHV", window_size=10, d_prime=1.0)
        
        # testing values
        try:
            self.assertEqual(reference, test)
            
        #TODO: why this test fail sometimes?
        except AssertionError as e:
            # sometimes this test can fail. In such case, i log the error
            logger.error(e)
            logger.error("Sometimes 'test_getLdId' fails. Maybe could be an ensembl transient problem?")
        
    def test_getLdPairwise(self):
        """Testing get LD pairwise GET method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/ld/human/pairwise/rs6792369/rs1042779?population_name=1000GENOMES:phase_3:KHV;r2=0.85' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getLdPairwise(species="human", id1="rs6792369", id2="rs1042779", population_name="1000GENOMES:phase_3:KHV", r2=0.85)
        
        # testing values
        try:
            self.assertEqual(reference, test)
            
        #TODO: why this test fail sometimes?
        except AssertionError as e:
            # sometimes this test can fail. In such case, i log the error
            logger.error(e)
            logger.error("Sometimes 'test_getLdPairwise' fails. Maybe could be an ensembl transient problem?")
        
    def test_getLdRegion(self):
        """Testing get LD region GET method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/ld/human/region/6:25837556..25843455/1000GENOMES:phase_3:KHV?r2=0.85:d_prime=1.0' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getLdRegion(species="human", region="6:25837556..25843455", population_name="1000GENOMES:phase_3:KHV", r2=0.85, d_prime=1.0)
        
        # testing values
        try:
            self.assertTrue(reference, test)
            
        #TODO: why this test fail sometimes?
        except AssertionError as e:
            # sometimes this test can fail. In such case, i log the error
            logger.error(e)
            logger.error("Sometimes 'test_getLdRegion' fails. Maybe could be an ensembl transient problem?")
    
class EnsemblRestLookUp(EnsemblRest):
    """A class to deal with ensemblrest LookUp methods"""

    def test_getLookupById(self):
        """Testing get lookup by id GET method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/lookup/id/ENSG00000157764?expand=1' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getLookupById(id='ENSG00000157764', expand=1)
        
        # testing values
        self.assertEqual(reference, test)
        
    def test_getLookupByMultipleIds(self):
        """Testing get lookup id POST method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/lookup/id' -H 'Content-type:application/json' \
-H 'Accept:application/json' -X POST -d '{ "ids" : ["ENSG00000157764", "ENSG00000248378" ] }'"""

        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getLookupByMultipleIds(ids=["ENSG00000157764", "ENSG00000248378" ])
        
        # testing values
        self.assertEqual(reference, test)
        
    def test_getLookupByMultipleIds_additional_arguments(self):
        """Testing get lookup id POST method with additional arguments"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/lookup/id?expand=1' -H 'Content-type:application/json' \
-H 'Accept:application/json' -X POST -d '{ "ids" : ["ENSG00000157764", "ENSG00000248378" ] }'"""

        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getLookupByMultipleIds(ids=["ENSG00000157764", "ENSG00000248378" ], expand=1)
        
        # testing values
        self.assertEqual(reference, test)
        
    def test_getLookupBySymbol(self):
        """Testing get lookup by species GET method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/lookup/symbol/homo_sapiens/BRCA2?expand=1' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getLookupBySymbol(species="homo_sapiens", symbol="BRCA2", expand=1)
        
        # testing values
        self.assertEqual(reference, test)
    
    def test_getLookupByMultipleSymbols(self):
        """Testing get lookup by species POST method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/lookup/symbol/homo_sapiens' -H 'Content-type:application/json' \
-H 'Accept:application/json' -X POST -d '{ "symbols" : ["BRCA2", "BRAF" ] }'"""

        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getLookupByMultipleSymbols(species="homo_sapiens", symbols=["BRCA2", "BRAF"])
        
        # testing values
        self.assertEqual(reference, test)
    
    def test_getLookupByMultipleSymbols_additional_arguments(self):
        """Testing get lookup by species POST method  with additional arguments"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/lookup/symbol/homo_sapiens?expand=1' -H 'Content-type:application/json' \
-H 'Accept:application/json' -X POST -d '{ "symbols" : ["BRCA2", "BRAF" ] }'"""

        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getLookupByMultipleSymbols(species="homo_sapiens", symbols=["BRCA2", "BRAF"], expand=1)
        
        # testing values
        self.assertEqual(reference, test)
        
    
class EnsemblRestMapping(EnsemblRest):
    """A class to deal with ensemblrest mapping methods"""

    def test_getMapCdnaToRegion(self):
        """Testing map CDNA to region GET method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/map/cdna/ENST00000288602/100..300?' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getMapCdnaToRegion(id='ENST00000288602', region='100..300')
        
        # testing values
        self.assertEqual(reference, test)
        
    def test_getMapCdsToRegion(self):
        """Testing map CDS to region GET method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/map/cds/ENST00000288602/1..1000?' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getMapCdsToRegion(id='ENST00000288602', region='1..1000')
        
        # testing values
        self.assertEqual(reference, test)
        
    def test_getMapAssemblyOneToTwo(self):
        """Testing converting coordinates between assemblies GET method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/map/human/GRCh37/X:1000000..1000100:1/GRCh38?' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getMapAssemblyOneToTwo(species='human', asm_one='GRCh37', region='X:1000000..1000100:1', asm_two='GRCh38')
        
        # testing values
        self.assertEqual(reference, test)
        
    def test_getMapTranslationToRegion(self):
        """Testing converting protein(traslation) to genomic coordinates GET method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/map/translation/ENSP00000288602/100..300?' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getMapTranslationToRegion(id='ENSP00000288602', region='100..300')
        
        # testing values
        self.assertEqual(reference, test)
    
    
class EnsemblRestOT(EnsemblRest):
    """A class to deal with ensemblrest ontologies and taxonomy methods"""

    def test_getAncestorsById(self):
        """Testing get ancestors by id GET method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/ontology/ancestors/GO:0005667?' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getAncestorsById(id='GO:0005667')
        
        # testing values
        self.assertEqual(reference, test)
        
    def test_getAncestorsChartById(self):
        """Testing get ancestors chart by id GET method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/ontology/ancestors/chart/GO:0005667?' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getAncestorsChartById(id='GO:0005667')
        
        # testing values
        self.assertEqual(reference, test)
        
    def test_getDescendantsById(self):
        """Testing get descendants by id GET method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/ontology/descendants/GO:0005667?' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getDescendantsById(id='GO:0005667')
        
        # testing values
        self.assertEqual(reference, test)
        
    def test_getOntologyById(self):
        """Test get ontology by id GET method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/ontology/id/GO:0005667?' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getOntologyById(id='GO:0005667')
        
        # testing values
        self.assertEqual(reference, test)
        
    def test_getOntologyByName(self):
        """Test get ontology by name GET method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/ontology/name/%s?' -H 'Content-type:application/json'""" % (
            six.moves.urllib.parse.quote("transcription factor complex")
        )
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getOntologyByName(name='transcription factor complex')
        
        # testing values
        self.assertEqual(reference, test)
        
    def test_getTaxonomyClassificationById(self):
        """Testing get taxonomy classification by id GET method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/taxonomy/classification/9606?' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getTaxonomyClassificationById(id='9606')
        
        # testing values
        try:
            self.assertTrue(reference, test)
            
        #TODO: why this test fail sometimes?
        except AssertionError as e:
            # sometimes this test can fail. In such case, i log the error
            logger.error(e)
            logger.error("Sometimes 'test_getTaxonomyClassificationById' fails. Maybe could be an ensembl transient problem?")
        
    def test_getTaxonomyById(self):
        """Testing get Taxonomy by id GET method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/taxonomy/id/9606?' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getTaxonomyById(id='9606')
        
        try:
            # testing values. Since json are nested dictionary and lists, and they are not hashable, I need to order list before
            # checking equality, and I need to ensure that dictionaries have the same keys and values
            self.assertTrue(compareDict(reference, test))
        
        # The transitory failure seems to be related to a misconfiguration of ensembl
        # rest service. In such cases is better to inform dev<at>ensembl.org and report
        # such issues
        except AssertionError as e:
            # sometimes this test can fail. In such case, i log the error
            logger.error(e)
            logger.error("Sometimes 'test_getTaxonomyById' fails. This could be a transitory problem on EnsEMBL REST service")
        
    def test_getTaxonomyByName(self):
        """Testing get taxonomy by name GET method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/taxonomy/name/Homo%25?' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getTaxonomyByName(name="Homo%25")
        
        # testing values. Since json are nested dictionary and lists, and they are not hashable, I need to order list before
        # checking equality, and I need to ensure that dictionaries have the same keys and values
        self.assertTrue(compareList(reference, test))


class EnsemblRestOverlap(EnsemblRest):
    """A class to deal with ensemblrest overlap methods"""

    def test_getOverlapById(self):
        """Testing get Overlap by ID GET method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/overlap/id/ENSG00000157764?feature=gene' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getOverlapById(id="ENSG00000157764", feature="gene")
        
        # testing values
        self.assertEqual(reference, test)
        
    def test_getOverlapByRegion(self):
        """Testing get Overlap by region GET method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/overlap/region/human/7:140424943-140624564?feature=gene;feature=transcript;feature=cds;feature=exon' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getOverlapByRegion(species="human", region="7:140424943-140624564", feature=["gene", "transcript", "cds", "exon"])
        
        # testing values
        self.assertEqual(reference, test)
        
    def test_getOverlapByTranslation(self):
        """Testing get Overlab by traslation GET method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/overlap/translation/ENSP00000288602?type=Superfamily' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getOverlapByTranslation(id="ENSP00000288602", type="SuperFamily")
        
        # testing values
        self.assertEqual(reference, test)

    
    # Regulation
    def test_getRegulatoryFeatureById(self):
        """Testing get regulatory Feature GET method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/regulatory/species/human/id/ENSR00000099113?' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getRegulatoryFeatureById(species="human", id="ENSR00000099113")
        
        # testing values
        self.assertEqual(reference, test)
        
    
class EnsemblRestSequence(EnsemblRest):
    """A class to deal with ensemblrest sequence methods"""

    def test_getSequenceById(self):
        """Testing get sequence by ID GET method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/sequence/id/CCDS5863.1?object_type=transcript;db_type=otherfeatures;type=cds;species=human' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getSequenceById(id='CCDS5863.1', object_type="transcript", db_type="otherfeatures", type="cds", species="human")
        
        # testing values
        self.assertEqual(reference, test)
        
    def test_getSequenceByMultipleIds(self):
        """Testing get sequence by ID POST method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/sequence/id' -H 'Content-type:application/json' \
-H 'Accept:application/json' -X POST -d '{ "ids" : ["ENSG00000157764", "ENSG00000248378" ] }'"""

        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getSequenceByMultipleIds(ids=["ENSG00000157764", "ENSG00000248378" ])
        
        # testing values
        self.assertEqual(reference, test)
        
    def test_getSequenceByMultipleIds_additional_arguments(self):
        """Testing getSequenceByMultipleIds with mask="soft" and expand_3prime=100"""

        curl_cmd = """curl 'http://rest.ensembl.org/sequence/id?mask=soft;expand_3prime=100' -H 'Content-type:application/json' \
-H 'Accept:application/json' -X POST -d '{ "ids" : ["ENSG00000157764", "ENSG00000248378" ] }'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getSequenceByMultipleIds(ids=["ENSG00000157764", "ENSG00000248378"], expand_3prime=100, mask="soft")
        
        # testing values
        self.assertEqual(reference, test)
        
    def test_getSequenceByRegion(self):
        """Testing get sequence by region GET method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/sequence/region/human/X:1000000..1000100:1?' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getSequenceByRegion(species='human', region='X:1000000..1000100:1')
        
        # testing values
        self.assertTrue(compareDict(reference, test))
        
    def test_getSequenceByMultipleRegions(self):
        """Testing get sequence by region POST method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/sequence/region/human' -H 'Content-type:application/json' \
-H 'Accept:application/json' -X POST -d '{ "regions" : ["X:1000000..1000100:1", "ABBA01004489.1:1..100"] }'"""

        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getSequenceByMultipleRegions(species="human", regions=["X:1000000..1000100:1", "ABBA01004489.1:1..100"])
        
        # testing values
        self.assertEqual(reference, test)
        
    def test_getSequenceByMultipleRegions_additional_arguments(self):
        """Testing get sequence by region POST method with mask="soft" and expand_3prime=100"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/sequence/region/human?mask=soft;expand_3prime=100' -H 'Content-type:application/json' -H 'Accept:application/json' -X POST -d '{ "regions" : ["X:1000000..1000100:1", "ABBA01004489.1:1..100"] }'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getSequenceByMultipleRegions(species="human", regions=["X:1000000..1000100:1", "ABBA01004489.1:1..100"], expand_3prime=100, mask="soft")
        
        # testing values
        self.assertEqual(reference, test)
        
        
class EnsemblRestHaplotype(EnsemblRest):
    """A class to deal with ensemblrest transcript haplotypes methods"""

    def test_getTranscripsHaplotypes(self):
        """Testing get transcripts Haplotypes GET method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/transcript_haplotypes/homo_sapiens/ENST00000288602?' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getTranscripsHaplotypes(species="homo_sapiens", id="ENST00000288602")
        
        # testing values
        self.assertEqual(reference, test)

    
class EnsemblRestVEP(EnsemblRest):
    """A class to deal with ensemblrest Variant Effect Predictor methods"""

    def test_getVariantConsequencesByHGVSnotation(self):
        """Testing get Variant Consequences by HFVS notation GET method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/vep/human/hgvs/AGT:c.803T>C?' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getVariantConsequencesByHGVSnotation(species="human", hgvs_notation="AGT:c.803T>C")
        
        # testing values
        self.assertEqual(reference, test)
        
    def test_getVariantConsequencesById(self):
        """Testing get variant Consequences by id GET method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/vep/human/id/COSM476?' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getVariantConsequencesById(species='human', id='COSM476')
        
        # testing values
        self.assertEqual(reference, test)
        
    def test_getVariantConsequencesByMultipleIds(self):
        """Testing get variant Consequences by id POST method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/vep/human/id' -H 'Content-type:application/json' \
-H 'Accept:application/json' -X POST -d '{ "ids" : ["rs56116432", "COSM476" ] }'"""

        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getVariantConsequencesByMultipleIds(species="human", ids=[ "rs56116432", "COSM476" ])
        
        # testing values
        self.assertEqual(reference, test)
        
    def test_getVariantConsequencesByMultipleIds_additional_arguments(self):
        """Testing get variant Consequences by id POST method using Blosum62=1, CSN=1"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/vep/human/id?Blosum62=1;CSN=1' -H 'Content-type:application/json' \
-H 'Accept:application/json' -X POST -d '{ "ids" : ["rs56116432", "COSM476" ] }'"""

        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getVariantConsequencesByMultipleIds(species="human", ids=[ "rs56116432", "COSM476" ], Blosum62=1, CSN=1)
        
        # testing values
        self.assertEqual(reference, test)
    
    def test_getVariantConsequencesByRegion(self):
        """Testing get variant consequences by Region GET method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/vep/human/region/9:22125503-22125502:1/C?' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getVariantConsequencesByRegion(species='human', region='9:22125503-22125502:1', allele='C')
        
        # testing values
        self.assertEqual(reference, test)
        
    def test_getVariantConsequencesByMultipleRegions(self):
        """Testing get variant consequences by Region POST method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/vep/homo_sapiens/region' -H 'Content-type:application/json' \
-H 'Accept:application/json' -X POST -d '{ "variants" : ["21 26960070 rs116645811 G A . . .", "21 26965148 rs1135638 G A . . ." ] }'"""

        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getVariantConsequencesByMultipleRegions(species="human", variants=["21 26960070 rs116645811 G A . . .", "21 26965148 rs1135638 G A . . ." ] )
        
        # testing values
        self.assertEqual(reference, test)
        
    def test_getVariantConsequencesByMultipleRegions_additional_arguments(self):
        """Testing get variant consequences by Region POST method Blosum62=1, CSN=1"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/vep/homo_sapiens/region?Blosum62=1;CSN=1' -H 'Content-type:application/json' \
-H 'Accept:application/json' -X POST -d '{ "variants" : ["21 26960070 rs116645811 G A . . .", "21 26965148 rs1135638 G A . . ." ] }'"""

        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getVariantConsequencesByMultipleRegions(species="human", variants=["21 26960070 rs116645811 G A . . .", "21 26965148 rs1135638 G A . . ." ], Blosum62=1, CSN=1 )
        
        # testing values
        self.assertEqual(reference, test)
    
    
class EnsemblRestVariation(EnsemblRest):
    """A class to deal with ensemblrest variation methods"""

    def test_getVariationById(self):
        """Testing get variation by id GET method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/variation/human/rs56116432?' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getVariationById(id="rs56116432", species="homo_sapiens")
        
        # testing values
        self.assertEqual(reference, test)
        
    def test_getVariationByMultipleIds(self):
        """Testing get variation by id POST method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/variation/homo_sapiens' -H 'Content-type:application/json' \
-H 'Accept:application/json' -X POST -d '{ "ids" : ["rs56116432", "COSM476" ] }'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getVariationByMultipleIds(ids=["rs56116432", "COSM476" ], species="homo_sapiens")
        
        # testing values
        self.assertEqual(reference, test)
        
    def test_getVariationByMultipleIds_additional_arguments(self):
        """Testing get variation by id POST method with genotypes=1"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/variation/homo_sapiens?genotypes=1' -H 'Content-type:application/json' \
-H 'Accept:application/json' -X POST -d '{ "ids" : ["rs56116432", "COSM476" ] }'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getVariationByMultipleIds(ids=["rs56116432", "COSM476" ], species="homo_sapiens", genotypes=1)
        
        # testing values
        self.assertEqual(reference, test)
    
    
class EnsemblRestVariationGA4GH(EnsemblRest):
    """A class to deal with ensemblrest variation GA4GH methods"""

    def test_searchGA4GHCallSet(self):
        """Testing GA4GH callset search POST method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/ga4gh/callsets/search' -H 'Content-type:application/json' \
-H 'Accept:application/json' -X POST -d '{ "variantSetId": 1, "pageSize": 2  }'"""

        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.searchGA4GHCallSet(variantSetId=1, pageSize=2)
        
        # testing values
        self.assertEqual(reference, test)
        
    def test_getGA4GHCallSetById(self):
        """Testing get GA4GH callset by Id GET method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/ga4gh/callsets/1:NA19777?' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getGA4GHCallSetById(id="1:NA19777")
        
        # testing values
        self.assertEqual(reference, test)
        
    def test_searchGA4GHDataset(self):
        """Testing GA4GH search dataset POST method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/ga4gh/datasets/search' -H 'Content-type:application/json' \
-H 'Accept:application/json' -X POST -d '{ "pageSize": 3 }'"""

        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.searchGA4GHDataset(pageSize=3)
        
        # testing values
        self.assertEqual(reference, test)
        
    def test_getGA4GHDatasetById(self):
        """Testing GA4GH get dataset by Id GET method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/ga4gh/datasets/6e340c4d1e333c7a676b1710d2e3953c?' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getGA4GHDatasetById(id="6e340c4d1e333c7a676b1710d2e3953c")
        
        # testing values
        self.assertEqual(reference, test)
        
    def test_getGA4GHVariantsById(self):
        """Testing GA4GH get variant by Id GET method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/ga4gh/variants/1:rs1333049?' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getGA4GHVariantsById(id="1:rs1333049")
        
        # testing values
        self.assertEqual(reference, test)
        
    def test_searchGA4GHVariants(self):
        """Testing GA4GH search variants POST method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/ga4gh/variants/search' -H 'Content-type:application/json' \
-H 'Accept:application/json' -X POST -d '{ "variantSetId": 1, "referenceName": 22,"start": 17190024 ,"end":  17671934 ,  "pageToken":"", "pageSize": 1 }'"""

        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.searchGA4GHVariants(variantSetId=1, referenceName=22, start=17190024, end=17671934, pageToken="", pageSize=1)
        
        # testing values
        self.assertEqual(reference, test)
        
    def test_searchGA4GHVariantsets(self):
        """Testing GA4GH search variantset POST method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/ga4gh/variantsets/search' -H 'Content-type:application/json' \
-H 'Accept:application/json' -X POST -d '{ "datasetId": "6e340c4d1e333c7a676b1710d2e3953c",    "pageToken": "", "pageSize": 2 }'"""

        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.searchGA4GHVariantsets(datasetId="6e340c4d1e333c7a676b1710d2e3953c", pageToken="", pageSize=2)
        
        # testing values
        self.assertEqual(reference, test)
        
    def test_getGA4GHVariantsetsById(self):
        """Testing GA4GH get variantset by Id GET method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/ga4gh/variantsets/1?' -H 'Content-type:application/json'"""

        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getGA4GHVariantsetsById(id=1)
        
        # testing values
        self.assertEqual(reference, test)
        
    def test_searchGA4GHReferences(self):
        """Testing GA4GH search references POST method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/ga4gh/references/search' -H 'Content-type:application/json' \
-H 'Accept:application/json' -X POST -d '{ "referenceSetId": "GRCh38", "pageSize": 10 }'"""

        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.searchGA4GHReferences(referenceSetId="GRCh38", pageSize=10)
        
        # testing values
        self.assertEqual(reference, test)
        
    def test_getGA4GHReferencesById(self):
        """Testing GA4GH get references by Id GET method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/ga4gh/references/9489ae7581e14efcad134f02afafe26c?' -H 'Content-type:application/json'"""

        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getGA4GHReferencesById(id="9489ae7581e14efcad134f02afafe26c")
        
        # testing values
        self.assertEqual(reference, test)
        
    def test_searchGA4GHReferenceSets(self):
        """Testing GA4GH search reference sets POST method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/ga4gh/referencesets/search' -H 'Content-type:application/json' \
-H 'Accept:application/json' -X POST -d '{   }'"""

        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.searchGA4GHReferenceSets()
        
        # testing values
        self.assertEqual(reference, test)
        
    def test_getGA4GHReferenceSetsById(self):
        """Testing GA4GH get reference set by Id GET method"""
        
        curl_cmd = """curl 'http://rest.ensembl.org/ga4gh/referencesets/GRCh38?' -H 'Content-type:application/json'"""

        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getGA4GHReferenceSetsById(id="GRCh38")
        
        # testing values
        self.assertEqual(reference, test)


class EnsemblGenomeRest(unittest.TestCase):
    """A class to test EnsemblGenomeRest methods"""
    
    def setUp(self):
        """Create a EnsemblRest object"""
        self.EnsEMBL = ensemblrest.EnsemblGenomeRest()
        
    def tearDown(self):
        """Sleep a while before doing next request"""
        time.sleep(WAIT)

    def test_getGeneFamilyById(self):
        """Testing genefamily by id GET method"""
        
        curl_cmd = """curl 'http://rest.ensemblgenomes.org/family/id/MF_01120?compara=bacteria' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getGeneFamilyById(id="MF_01120", compara="bacteria")
        
        # testing values
        self.assertEqual(reference, test)
        
    def test_getGeneFamilyMemberById(self):
        """Testing genefamily member by id GET method"""
        
        curl_cmd = """curl 'http://rest.ensemblgenomes.org/family/member/id/STK_01900?compara=bacteria' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getGeneFamilyMemberById(id="STK_01900", compara="bacteria")
        
        # testing values
        self.assertEqual(reference, test)
        
    def test_getGeneFamilyMemberBySymbol(self):
        """Testing genefamily member by symbol"""
        
        curl_cmd = """curl 'http://rest.ensemblgenomes.org/family/member/symbol/sulfolobus_tokodaii_str_7/lysK?compara=bacteria' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getGeneFamilyMemberBySymbol(symbol="lysK", species="sulfolobus_tokodaii_str_7", compara="bacteria")
        
        # testing values
        self.assertEqual(reference, test)
        
    def test_getInfoEgVersion(self):
        """Testing EgVersion GET method"""
        
        curl_cmd = """curl 'http://rest.ensemblgenomes.org/info/eg_version?' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getInfoEgVersion()
        
        # testing values
        self.assertEqual(reference, test)
        
    def test_getInfoDivisions(self):
        """Testing Info Divisions GET method"""
        
        curl_cmd = """curl 'http://rest.ensemblgenomes.org/info/divisions?' -H 'Content-type:application/json'"""

        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getInfoDivisions()
        
        # testing values
        self.assertEqual(reference, test)
        
    def test_getInfoGenomesByName(self):
        """Testing Info Genomes by Name GET method"""
        
        curl_cmd = """curl 'http://rest.ensemblgenomes.org/info/genomes/campylobacter_jejuni_subsp_jejuni_bh_01_0142?' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getInfoGenomesByName(name="campylobacter_jejuni_subsp_jejuni_bh_01_0142")
        
        # testing values
        self.assertEqual(reference, test)
        
    def test_getInfoGenomesByAccession(self): 
        """Testing Info Genomes by Accession GET method"""
        
        curl_cmd = """curl 'http://rest.ensemblgenomes.org/info/genomes/accession/U00096?' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getInfoGenomesByAccession(division="U00096")
        
        # testing values
        self.assertEqual(reference, test)
        
    def test_getInfoGenomesByAssembly(self):
        """Testing Info Genomes by Assembly GET method"""
        
        curl_cmd = """curl 'http://rest.ensemblgenomes.org/info/genomes/assembly/GCA_000005845?' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getInfoGenomesByAssembly(division="GCA_000005845")
        
        # testing values
        self.assertEqual(reference, test)
    
    def test_getInfoGenomesByDivision(self):
        """Testing Info Genomes by Division GET method"""
        
        curl_cmd = """curl 'http://rest.ensemblgenomes.org/info/genomes/division/EnsemblPlants?' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getInfoGenomesByDivision(division="EnsemblPlants")
        
        # testing values
        self.assertEqual(reference, test)
        
    def test_getInfoGenomesByTaxonomy(self):
        """Testing Info Genomes by Taxonomy GET method"""
        
        curl_cmd = """curl 'http://rest.ensemblgenomes.org/info/genomes/taxonomy/Arabidopsis?' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getInfoGenomesByTaxonomy(division="Arabidopsis")
        
        # testing values
        self.assertEqual(reference, test)
        
    def test_getLookupByGenomeName(self):
        """Testing Lookup by genome name GET method"""
        
        curl_cmd = """curl 'http://rest.ensemblgenomes.org/lookup/genome/campylobacter_jejuni_subsp_jejuni_bh_01_0142?' -H 'Content-type:application/json'"""
        
        # execute the curl cmd an get data as a dictionary
        reference = jsonFromCurl(curl_cmd)
      
        # execute EnsemblRest function
        test = self.EnsEMBL.getLookupByGenomeName(name="campylobacter_jejuni_subsp_jejuni_bh_01_0142")
        
        # testing values
        self.assertEqual(reference, test)

if __name__ == "__main__":
    unittest.main()

    


"""

    This file is part of pyEnsemblRest.
    Copyright (C) 2013-2016, Steve Moss

    pyEnsemblRest is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    pyEnsemblRest is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with pyEnsemblRest.  If not, see <http://www.gnu.org/licenses/>.

    Implements custom exceptions for the EnsEMBL REST API
    
"""
from .ensembl_config import ensembl_http_status_codes


class EnsemblRestError(Exception):
    """
        Generic error class, catch-all for most EnsemblRest issues.
        Special cases are handled by EnsemblRestRateLimitError and EnsemblRestServiceUnavailable.
    """
    def __init__(self, msg, error_code=None, rate_reset=None, rate_limit=None, rate_remaining=None, retry_after=None):
        self.error_code = error_code

        if error_code is not None and error_code in ensembl_http_status_codes:
            msg = 'EnsEMBL REST API returned a %s (%s): %s' % \
                    (error_code, ensembl_http_status_codes[error_code][0], msg)

        super(EnsemblRestError, self).__init__(msg)

    @property
    def msg(self):
        return self.args[0]


class EnsemblRestRateLimitError(EnsemblRestError):
    """
        Raised when you've hit a rate limit.
        The amount of seconds to retry your request in will be appended to the message.
    """
    def __init__(self, msg, error_code=None, rate_reset=None, rate_limit=None, rate_remaining=None, retry_after=None):
        if isinstance(retry_after, float):
            msg = '%s (Rate limit hit:  Retry after %d seconds)' % (msg, retry_after)
            
        EnsemblRestError.__init__(self, msg, error_code=error_code)


class EnsemblRestServiceUnavailable(EnsemblRestError):
    """
        Raised when the service is down.
    """
    pass



"""

    This file is part of pyEnsemblRest.
    Copyright (C) 2013-2016, Steve Moss

    pyEnsemblRest is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    pyEnsemblRest is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with pyEnsemblRest.  If not, see <http://www.gnu.org/licenses/>.

    Configuration information for the EnsEMBL REST API
    
"""

# import pyensemblrest version
import six
from . import __version__

# set urls
ensembl_default_url = 'http://rest.ensembl.org'
ensembl_genomes_url = 'http://rest.ensemblgenomes.org'

# ensembl api lookup table. Specify here function common to ensembl and ensemblgenomes
# REST server
ensembl_api_table = {
    # Archive
    'getArchiveById': {
        'doc': 'Uses the given identifier to return the archived sequence',
        'url': '/archive/id/{{id}}',
        'method': 'GET',
        'content_type': 'application/json'
    },
    'getArchiveByMultipleIds': {
        'doc': 'Retrieve the archived sequence for a set of identifiers',
        'url': '/archive/id',
        'method': 'POST',
        'content_type': 'application/json',
        'post_parameters': ['id']
    },
    

    # Comparative Genomics
    'getGeneTreeById': {
        'doc': 'Retrieves a gene tree dump for a gene tree stable identifier',
        'url': '/genetree/id/{{id}}',
        'method': 'GET',
        'content_type': 'text/x-phyloxml+xml'
    },
    'getGeneTreeMemberById': {
        'doc': 'Retrieves a gene tree that contains the stable identifier',
        'url': '/genetree/member/id/{{id}}',
        'method': 'GET',
        'content_type': 'text/x-phyloxml+xml'
    },
    'getGeneTreeMemberBySymbol': {
        'doc': 'Retrieves a gene tree containing the gene identified by a symbol',
        'url': '/genetree/member/symbol/{{species}}/{{symbol}}',
        'method': 'GET',
        'content_type': 'text/x-phyloxml+xml'
    },
    'getAlignmentByRegion': {
        'doc': 'Retrieves genomic alignments as separate blocks based on a region and species',
        'url': '/alignment/region/{{species}}/{{region}}',
        'method': 'GET',
        'content_type': 'application/json'
    },
    'getHomologyById': {
        'doc': 'Retrieves homology information (orthologs) by Ensembl gene id',
        'url': '/homology/id/{{id}}',
        'method': 'GET',
        'content_type': 'application/json'
    },
    'getHomologyBySymbol': {
        'doc': 'Retrieves homology information (orthologs) by symbol',
        'url': '/homology/symbol/{{species}}/{{symbol}}',
        'method': 'GET',
        'content_type': 'application/json'
    },
    

    # Cross References
    'getXrefsBySymbol': {
        'doc': """Looks up an external symbol and returns all Ensembl objects linked to it. """
               """This can be a display name for a gene/transcript/translation, """
               """a synonym or an externally linked reference. """
               """If a gene's transcript is linked to the supplied symbol """
               """the service will return both gene and transcript (it supports transient links).""",
        'url': '/xrefs/symbol/{{species}}/{{symbol}}',
        'method': 'GET',
        'content_type': 'application/json'
    },
    'getXrefsById': {
        'doc': "Perform lookups of Ensembl Identifiers and retrieve their external references in other databases",
        'url': '/xrefs/id/{{id}}',
        'method': 'GET',
        'content_type': 'application/json'
    },
    'getXrefsByName': {
        'doc': "Performs a lookup based upon the primary accession or display label of an external reference "
               "and returning the information we hold about the entry",
        'url': '/xrefs/name/{{species}}/{{name}}',
        'method': 'GET',
        'content_type': 'application/json'
    },
    

    # Information
    'getInfoAnalysis': {
        'doc': 'List the names of analyses involved in generating Ensembl data',
        'url': '/info/analysis/{{species}}',
        'method': 'GET',
        'content_type': 'application/json'
    },
    'getInfoAssembly': {
        'doc': 'List the currently available assemblies for a species',
        'url': '/info/assembly/{{species}}',
        'method': 'GET',
        'content_type': 'application/json'
    },
    'getInfoAssemblyRegion': {
        'doc': 'Returns information about the specified toplevel sequence region for the given species',
        'url': '/info/assembly/{{species}}/{{region_name}}',
        'method': 'GET',
        'content_type': 'application/json'
    },
    'getInfoBiotypes': {
        'doc': 'List the functional classifications of gene models that Ensembl associates with a particular species. '
               'Useful for restricting the type of genes/transcripts retrieved by other endpoints',
        'url': '/info/biotypes/{{species}}',
        'method': 'GET',
        'content_type': 'application/json',
    },
    'getInfoComparaMethods': {
        'doc': 'List all compara analyses available (an analysis defines the type of comparative data)',
        'url': '/info/compara/methods',
        'method': 'GET',
        'content_type': 'application/json',
    },
    'getInfoComparaSpeciesSets': {
        'doc': 'List all collections of species analysed with the specified compara method',
        'url': '/info/compara/species_sets/{{methods}}',
        'method': 'GET',
        'content_type': 'application/json',
    },
    'getInfoComparas': {
        'doc': 'Lists all available comparative genomics databases and their data release',
        'url': '/info/comparas',
        'method': 'GET',
        'content_type': 'application/json'
    },
    'getInfoData': {
        'doc': 'Shows the data releases available on this REST server. '
               'May return more than one release (unfrequent non-standard Ensembl configuration)',
        'url': '/info/data',
        'method': 'GET',
        'content_type': 'application/json'
    },
    'getInfoExternalDbs': {
        'doc': 'Lists all available external sources for a species',
        'url': '/info/external_dbs/{{species}}',
        'method': 'GET',
        'content_type': 'application/json',
    },    
    'getInfoPing': {
        'doc': 'Checks if the service is alive',
        'url': '/info/ping',
        'method': 'GET',
        'content_type': 'application/json'
    },
    'getInfoRest': {
        'doc': 'Shows the current version of the Ensembl REST API',
        'url': '/info/rest',
        'method': 'GET',
        'content_type': 'application/json'
    },
    'getInfoSoftware': {
        'doc': 'Shows the current version of the Ensembl API used by the REST server',
        'url': '/info/software',
        'method': 'GET',
        'content_type': 'application/json'
    },
    'getInfoSpecies': {
        'doc': 'Lists all available species, their aliases, available adaptor groups and data release',
        'url': '/info/species',
        'method': 'GET',
        'content_type': 'application/json'
    },
    'getInfoVariation': {
        'doc': 'List the variation sources used in Ensembl for a species',
        'url': '/info/variation/{{species}}',
        'method': 'GET',
        # Default ensembl content type
    },
    'getInfoVariationPopulations': {
        'doc': 'List all populations for a species',
        'url': '/info/variation/populations/{{species}}',
        'method': 'GET',
        # Default ensembl content type
    },
    
    
    # Linkage Disequilibrium
    'getLdId': {
        'doc': 'Computes and returns LD values between the given variant and all other variants '
               'in a window centered around the given variant. The window size is set to 500 kb',
        'url': '/ld/{{species}}/{{id}}/{{population_name}}',
        'method': 'GET',
        # Default ensembl content type
    },
    'getLdPairwise': {
        'doc': 'Computes and returns LD values between the given variants',
        'url': '/ld/{{species}}/pairwise/{{id1}}/{{id2}}',
        'method': 'GET',
        # Default ensembl content type
    },
    'getLdRegion': {
        'doc': 'Computes and returns LD values between all pairs of variants in the defined region',
        'url': '/ld/{{species}}/region/{{region}}/{{population_name}}',
        'method': 'GET',
        # Default ensembl content type
    },
    
    
    # Lookup
    'getLookupById': {
        'doc': 'Find the species and database for a single identifier',
        'url': '/lookup/id/{{id}}',
        'method': 'GET',
        'content_type': 'application/json'
    },
    'getLookupByMultipleIds': {
        'doc': 'Find the species and database for several identifiers. '
               'IDs that are not found are returned with no data',
        'url': '/lookup/id',
        'method': 'POST',
        'content_type': 'application/json',
        'post_parameters': ['ids']
    },
    'getLookupBySymbol': {
        'doc': 'Find the species and database for a symbol in a linked external database',
        'url': '/lookup/symbol/{{species}}/{{symbol}}',
        'method': 'GET',
        'content_type': 'application/json',
    },
    'getLookupByMultipleSymbols': {
        'doc': 'Find the species and database for a set of symbols in a linked external database. '
               'Unknown symbols are omitted from the response',
        'url': '/lookup/symbol/{{species}}',
        'method': 'POST',
        'content_type': 'application/json',
        'post_parameters': ['symbols']
    },


    # Mapping
    'getMapCdnaToRegion': {
        'doc': 'Convert from cDNA coordinates to genomic coordinates. '
               'Output reflects forward orientation coordinates as returned from the Ensembl API',
        'url': '/map/cdna/{{id}}/{{region}}',
        'method': 'GET',
        'content_type': 'application/json'
    },
    'getMapCdsToRegion': {
        'doc': 'Convert from CDS coordinates to genomic coordinates. '
               'Output reflects forward orientation coordinates as returned from the Ensembl API',
        'url': '/map/cds/{{id}}/{{region}}',
        'method': 'GET',
        'content_type': 'application/json'
    },
    'getMapAssemblyOneToTwo': {
        'doc': 'Convert the co-ordinates of one assembly to another',
        'url': '/map/{{species}}/{{asm_one}}/{{region}}/{{asm_two}}',
        'method': 'GET',
        'content_type': 'application/json'
    },
    'getMapTranslationToRegion': {
        'doc': 'Convert from protein (translation) coordinates to genomic coordinates. '
               'Output reflects forward orientation coordinates as returned from the Ensembl API',
        'url': '/map/translation/{{id}}/{{region}}',
        'method': 'GET',
        'content_type': 'application/json'
    },

    # Ontologies and Taxonomy
    'getAncestorsById': {
        'doc': 'Reconstruct the entire ancestry of a term from is_a and part_of relationships',
        'url': '/ontology/ancestors/{{id}}',
        'method': 'GET',
        'content_type': 'application/json'
    },
    'getAncestorsChartById': {
        'doc': 'Reconstruct the entire ancestry of a term from is_a and part_of relationships',
        'url': '/ontology/ancestors/chart/{{id}}',
        'method': 'GET',
        'content_type': 'application/json'
    },
    'getDescendantsById': {
        'doc': 'Find all the terms descended from a given term. '
               'By default searches are conducted within the namespace of the given identifier',
        'url': '/ontology/descendants/{{id}}',
        'method': 'GET',
        'content_type': 'application/json'
    },
    'getOntologyById': {
        'doc': 'Search for an ontological term by its namespaced identifier',
        'url': '/ontology/id/{{id}}',
        'method': 'GET',
        'content_type': 'application/json'
    },
    'getOntologyByName': {
        'doc': 'Search for a list of ontological terms by their name',
        'url': '/ontology/name/{{name}}',
        'method': 'GET',
        'content_type': 'application/json'
    },
    'getTaxonomyClassificationById': {
        'doc': 'Return the taxonomic classification of a taxon node',
        'url': '/taxonomy/classification/{{id}}',
        'method': 'GET',
        'content_type': 'application/json'
    },
    'getTaxonomyById': {
        'doc': 'Search for a taxonomic term by its identifier or name',
        'url': '/taxonomy/id/{{id}}',
        'method': 'GET',
        'content_type': 'application/json'
    },
    'getTaxonomyByName': {
        'doc': 'Search for a taxonomic id by a non-scientific name',
        'url': '/taxonomy/name/{{name}}',
        'method': 'GET',
        'content_type': 'application/json',
    },

    # Overlap
    'getOverlapById': {
        'doc': 'Retrieves features (e.g. genes, transcripts, variations etc.) '
               'that overlap a region defined by the given identifier',
        'url': '/overlap/id/{{id}}',
        'method': 'GET',
        'content_type': 'application/json',
    },
    'getOverlapByRegion': {
        'doc': 'Retrieves multiple types of features for a given region',
        'url': '/overlap/region/{{species}}/{{region}}',
        'method': 'GET',
        'content_type': 'application/json',
    },
    'getOverlapByTranslation': {
        'doc': 'Retrieve features related to a specific Translation '
               'as described by its stable ID (e.g. domains, variations)',
        'url': '/overlap/translation/{{id}}',
        'method': 'GET',
        'content_type': 'application/json',
    },

    
    # Regulation
    'getRegulatoryFeatureById': {
        'doc': 'Returns a RegulatoryFeature given its stable ID (e.g. ENSR00001348195)',
        'url': '/regulatory/species/{{species}}/id/{{id}}',
        'method': 'GET',
        'content_type': 'application/json'
    },


    # Sequences
    'getSequenceById': {
        'doc': 'Request multiple types of sequence by stable identifier',
        'url': '/sequence/id/{{id}}',
        'method': 'GET',
        'content_type': 'application/json'
    },
    'getSequenceByMultipleIds': {
        'doc': 'Request multiple types of sequence by a stable identifier list',
        'url': '/sequence/id',
        'method': 'POST',
        'content_type': 'application/json',
        'post_parameters': ['ids']
    },
    'getSequenceByRegion': {
        'doc': 'Returns the genomic sequence of the specified region of the given species',
        'url': '/sequence/region/{{species}}/{{region}}',
        'method': 'GET',
        'content_type': 'application/json'
    },
    'getSequenceByMultipleRegions': {
        'doc': 'Request multiple types of sequence by a list of regions',
        'url': '/sequence/region/{{species}}',
        'method': 'POST',
        'content_type': 'application/json',
        'post_parameters': ['regions']
    },


    # Transcript Haplotypes
    'getTranscripsHaplotypes': {
        'doc': 'Computes observed transcript haplotype sequences based on phased genotype data',
        'url': '/transcript_haplotypes/{{species}}/{{id}}',
        'method': 'GET',
        # Default ensembl content type
    },


    # VEP
    'getVariantConsequencesByHGVSnotation': {
        'doc': 'Request multiple types of sequence by a list of regions',
        'url': '/vep/{{species}}/hgvs/{{hgvs_notation}}',
        'method': 'GET',
        'content_type': 'application/json'
    },
    'getVariantConsequencesById': {
        'doc': 'Fetch variant consequences based on a variation identifier',
        'url': '/vep/{{species}}/id/{{id}}',
        'method': 'GET',
        'content_type': 'application/json'
    },
    'getVariantConsequencesByMultipleIds': {
        'doc': 'Fetch variant consequences for multiple ids',
        'url': '/vep/{{species}}/id',
        'method': 'POST',
        'content_type': 'application/json',
        'post_parameters': ['ids']
        
    },
    'getVariantConsequencesByRegion': {
        'doc': 'Fetch variant consequences',
        'url': '/vep/{{species}}/region/{{region}}/{{allele}}',
        'method': 'GET',
        'content_type': 'application/json'
    },
    'getVariantConsequencesByMultipleRegions': {
        'doc': 'Fetch variant consequences for multiple regions',
        'url': '/vep/{{species}}/region',
        'method': 'POST',
        'content_type': 'application/json',
        'post_parameters': ['variants']
    },
    
    
    # Variation
    'getVariationById': {
        'doc': 'Uses a variation identifier (e.g. rsID) to return the variation features',
        'url': '/variation/{{species}}/{{id}}',
        'method': 'GET',
        'content_type': 'application/json'
    },
    'getVariationByMultipleIds': {
        'doc': 'Uses a list of variant identifiers (e.g. rsID) to return the variation features '
               'including optional genotype, phenotype and population data',
        'url': '/variation/{{species}}',
        'method': 'POST',
        'content_type': 'application/json',
        'post_parameters': ['ids']
    },
    
    
    # Variation GA4GH
    'searchGA4GHCallSet': {
        'doc': 'Return a list of sets of genotype calls for specific samples in GA4GH format',
        'url': '/ga4gh/callsets/search',
        'method': 'POST',
        # Default ensembl content type
        'post_parameters': ["variantSetId", "name", "pageToken", "pageSize"],
    },
    'getGA4GHCallSetById': {
        'doc': "Return the GA4GH record for a specific CallSet given its identifier",
        'url': "/ga4gh/callsets/{{id}}",
        'method': 'GET',
        # Default ensembl content type
    },
    'searchGA4GHDataset': {
        'doc': 'Return a list of datasets in GA4GH format',
        'url': '/ga4gh/datasets/search',
        'method': 'POST',
        # Default ensembl content type
        'post_parameters': ["pageToken", "pageSize"],
    },
    'getGA4GHDatasetById': {
        'doc': "Return the GA4GH record for a specific dataset given its identifier",
        'url': "/ga4gh/datasets/{{id}}",
        'method': 'GET',
        # Default ensembl content type
    },
    'getGA4GHVariantsById': {
        'doc': "Return the GA4GH record for a specific variant given its identifier",
        'url': "/ga4gh/variants/{{id}}",
        'method': 'GET',
        # Default ensembl content type
    },
    'searchGA4GHVariants': {
        'doc': 'Return variant call information in GA4GH format for a region on a reference sequence',
        'url': '/ga4gh/variants/search',
        'method': 'POST',
        # Default ensembl content type
        'post_parameters': ["variantSetId", "callSetIds", "referenceName", "start", "end", "pageToken", "pageSize"],
    },
    'searchGA4GHVariantsets': {
        'doc': 'Return a list of variant sets in GA4GH format',
        'url': '/ga4gh/variantsets/search',
        'method': 'POST',
        # Default ensembl content type
        'post_parameters': ["datasetId", "pageToken", "pageSize"],
    },
    'getGA4GHVariantsetsById': {
        'doc': "Return the GA4GH record for a specific VariantSet given its identifier",
        'url': "/ga4gh/variantsets/{{id}}",
        'method': 'GET',
        # Default ensembl content type
    },
    'searchGA4GHReferences': {
        'doc': 'Return a list of reference sequences in GA4GH format',
        'url': '/ga4gh/references/search',
        'method': 'POST',
        # Default ensembl content type
        'post_parameters': ["referenceSetId", "md5checksum", "accession", "pageToken", "pageSize"],
    },
    'getGA4GHReferencesById': {
        'doc': "Return data for a specific reference in GA4GH format by id",
        'url': "/ga4gh/references/{{id}}",
        'method': 'GET',
        # Default ensembl content type
    },
    'searchGA4GHReferenceSets': {
        'doc': 'Return a list of reference sets in GA4GH format',
        'url': '/ga4gh/referencesets/search',
        'method': 'POST',
        # Default ensembl content type
        'post_parameters': ["accession", "pageToken", "pageSize"],
    },
    'getGA4GHReferenceSetsById': {
        'doc': "Return data for a specific reference set in GA4GH format",
        'url': "/ga4gh/referencesets/{{id}}",
        'method': 'GET',
        # Default ensembl content type
    },
}

# ensembl api lookup table. Specific of EnsEMBL genomes REST server. Specify here 
# functions defined in EnsemblGenomes REST server
ensemblgenomes_api_table = {
    # Comparative Genomics
    'getGeneFamilyById': {
        'doc': 'Retrieves gene family information by ID',
        'url': '/family/id/{{id}}',
        'method': 'GET',
        'content_type': 'application/json'
    },
    'getGeneFamilyMemberById': {
        'doc': 'Retrieves gene families to which a gene belongs',
        'url': '/family/member/id/{{id}}',
        'method': 'GET',
        'content_type': 'application/json'
    },
    'getGeneFamilyMemberBySymbol': {
        'doc': 'Retrieves gene families to which a gene belongs',
        'url': '/family/member/symbol/{{species}}/{{symbol}}',
        'method': 'GET',
        'content_type': 'application/json'
    },
    
    
    # Info
    'getInfoEgVersion': {
        'doc': 'Returns the Ensembl Genomes version of the databases backing this service',
        'url': '/info/eg_version',
        'method': 'GET',
        'content_type': 'application/json',
    },
    'getInfoDivisions': {
        'doc': 'Get list of all Ensembl divisions for which information is available',
        'url': '/info/divisions',
        'method': 'GET',
        'content_type': 'application/json',
    },    
    'getInfoGenomesByName': {
        'doc': 'Find information about a given genome',
        'url': '/info/genomes/{{name}}',
        'method': 'GET',
        'content_type': 'application/json',
    },    
    'getInfoGenomes': {
        'doc': 'Find information about all genomes. Response may be very large',
        'url': '/info/genomes',
        'method': 'GET',
        'content_type': 'application/json',
    },    
    'getInfoGenomesByAccession': {
        'doc': 'Find information about genomes containing a specified INSDC accession',
        'url': '/info/genomes/accession/{{division}}',
        'method': 'GET',
        'content_type': 'application/json',
    },    
    'getInfoGenomesByAssembly': {
        'doc': 'Find information about a genome with a specified assembly',
        'url': '/info/genomes/assembly/{{division}}',
        'method': 'GET',
        'content_type': 'application/json',
    },    
    'getInfoGenomesByDivision': {
        'doc': 'Find information about all genomes in a given division. May be large for Ensembl Bacteria',
        'url': '/info/genomes/division/{{division}}',
        'method': 'GET',
        'content_type': 'application/json',
    },    
    'getInfoGenomesByTaxonomy': {
        'doc': 'Find information about all genomes beneath a given node of the taxonomy',
        'url': '/info/genomes/taxonomy/{{division}}',
        'method': 'GET',
        'content_type': 'application/json',
    },

    
    # Lookup
    'getLookupByGenomeName': {
        'doc': 'Query for a named genome and retrieve the gene models',
        'url': '/lookup/genome/{{name}}',
        'method': 'GET',
        'content_type': 'application/json',
    },
}

# Add default ensemblrest methods to ensembgenome methods
for key, value in six.iteritems(ensembl_api_table):
    ensemblgenomes_api_table[key] = value

# http status codes
ensembl_http_status_codes = {
    200: ('OK', 'Request was a success. Only process data from the service when you receive this code'),
    400: ('Bad Request', 'Occurs during exceptional circumstances such as the service is unable to find an ID. '
                         'Check if the response Content-type was JSON. '
                         'If so the JSON object is an exception hash with the message keyed under error'),
    404: ('Not Found', 'Indicates a badly formatted request. Check your URL'),
    415: ('Unsupported Media Type', 'The server is refusing to service the request '
                                    'because the entity of the request is in a format not supported '
                                    'by the requested resource for the requested method'),
    429: ('Too Many Requests', 'You have been rate-limited; wait and retry. '
                               'The headers X-RateLimit-Reset, X-RateLimit-Limit and X-RateLimit-Remaining will '
                               'inform you of how long you have until your limit is reset and what that limit was. '
                               'If you get this response and have not exceeded your limit '
                               'then check if you have made too many requests per second.'),
    500: ('Internal Server Error', 'This error is not documented. Maybe there is an error in user input or '
                                   'REST server could have problems. Try to do the query with curl. '
                                   'If your data input and query are correct, contact ensembl team'),
    503: ('Service Unavailable', 'The service is temporarily down; retry after a pause'),
}

# set user agent
ensembl_user_agent = 'pyEnsemblRest v' + __version__
ensembl_header = {'User-Agent': ensembl_user_agent}
ensembl_content_type = 'application/json'

# define known errors
ensembl_known_errors = [
    "something bad has happened",
    "Something went wrong while fetching from LDFeatureContainerAdaptor",
    "%s timeout" % ensembl_user_agent
]



"""

    This file is part of pyEnsemblRest.
    Copyright (C) 2013-2016, Steve Moss

    pyEnsemblRest is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    pyEnsemblRest is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with pyEnsemblRest.  If not, see <http://www.gnu.org/licenses/>.

    EnsemblRest is a library for Python that wrap the EnsEMBL REST API.
    It simplifies all the API endpoints by abstracting them away from the
    end user and thus also ensures that an amendments to the library and/or
    EnsEMBL REST API won't cause major problems to the end user.

    Any questions, comments or issues can be addressed to gawbul@gmail.com.
    
"""

# import system modules
import re
import json
import time
import logging
import requests
from collections import namedtuple

# import ensemblrest modules
from .ensembl_config import ensembl_default_url, ensembl_genomes_url, ensembl_api_table, ensemblgenomes_api_table,\
    ensembl_http_status_codes, ensembl_header, ensembl_content_type, ensembl_known_errors, ensembl_user_agent
from .exceptions import EnsemblRestError, EnsemblRestRateLimitError, EnsemblRestServiceUnavailable

# Logger instance
logger = logging.getLogger(__name__)


# EnsEMBL REST API object
class EnsemblRest(object):
    # class initialisation function
    def __init__(self, api_table=ensembl_api_table, **kwargs):
        # read args variable into object as session_args
        self.session_args = kwargs or {}
        
        #In order to rate limiting the requests, like https://github.com/Ensembl/ensembl-rest/wiki/Example-Python-Client
        self.reqs_per_sec = 15
        self.req_count = 0
        self.last_req = 0
        self.wall_time = 1
        
        # get rate limit parameters, if provided
        self.rate_reset = None
        self.rate_limit = None
        self.rate_remaining = None
        self.retry_after = None
        
        # to record the last parameters used (in order to redo the query with an ensembl known error)
        self.last_url = None
        self.last_headers = {}
        self.last_params = {}
        self.last_data = {}
        self.last_method = None
        self.last_attempt = 0
        self.last_response = None
        
        # the maximum number of attempts
        self.max_attempts = 5
        
        # setting a timeout
        self.timeout = 60
        
        # set default values if those values are not provided
        self.__set_default()
        
        # setup requests session
        self.session = requests.Session()
        
        # update headers
        self.__update_headers()

        # add class methods relying api_table
        self.__add_methods(api_table)
            
    def __set_default(self):
        """Set default values"""
        
        # initialise default values
        default_base_url = ensembl_default_url
        default_headers = ensembl_header
        default_content_type = ensembl_content_type
        default_proxies = {}
        
        if 'base_url' not in self.session_args:
            self.session_args['base_url'] = default_base_url
            
        if 'headers' not in self.session_args:
            self.session_args['headers'] = default_headers
        
        if 'User-Agent' not in self.session_args['headers']:
            self.session_args['headers'].update(default_headers)
                            
        if 'Content-Type' not in self.session_args['headers']:
            self.session_args['headers']['Content-Type'] = default_content_type
            
        if 'proxies' not in self.session_args:
            self.session_args['proxies'] = default_proxies
    
    def __update_headers(self):
        """Update headers"""
        
        # update requests client with arguments
        client_args_copy = self.session_args.copy()
        for key, val in client_args_copy.items():
            if key in ('base_url', 'proxies'):
                setattr(self.session, key, val)
                self.session_args.pop(key)
        
        # update headers as already exist within client
        self.session.headers.update(self.session_args.pop('headers'))
        
    def __add_methods(self, api_table):
        """Add methods to class object"""
        
        # iterate over api_table keys and add key to class namespace
        for fun_name in api_table.keys():
            #setattr(self, key, self.register_api_func(key))
            #Not as a class attribute, but a class method
            self.__dict__[fun_name] = self.register_api_func(fun_name, api_table)
            
            #Set __doc__ for generic class method
            if "doc" in api_table[fun_name]:
                self.__dict__[fun_name].__doc__ = api_table[fun_name]["doc"]
            
            #add function name to the class methods
            self.__dict__[fun_name].__name__ = fun_name
        
    # dynamic api registration function
    def register_api_func(self, api_call, api_table):
        return lambda **kwargs: self.call_api_func(api_call, api_table, **kwargs)

    @staticmethod
    def __check_params(func, kwargs):
        """Check for mandatory parameters"""
        
        #Verify required variables and raise an Exception if needed
        mandatory_params = re.findall('\{\{(?P<m>[a-zA-Z1-9_]+)\}\}', func['url'])
        
        for param in mandatory_params:
            if param not in kwargs:
                logger.critical("'%s' param not specified. Mandatory params are %s" % (param, mandatory_params))
                raise Exception("mandatory param '%s' not specified" % param)
            else:
                logger.debug("Mandatory param %s found" % param)
                
        return mandatory_params

    # dynamic api call function
    def call_api_func(self, api_call, api_table, **kwargs):
        # build url from api_table kwargs
        func = api_table[api_call]
        
        # check mandatory params
        mandatory_params = self.__check_params(func, kwargs)
        
        # resolving urls
        url = re.sub('\{\{(?P<m>[a-zA-Z1-9_]+)\}\}', lambda m: "%s" % kwargs.get(m.group(1)),
                     self.session.base_url + func['url'])
        
        # debug
        logger.debug("Resolved url: '%s'" % url)
        
        # Now I have to remove mandatory params from kwargs        
        for param in mandatory_params:
            del(kwargs[param])
            
        # Initialize with the ensembl default content type
        content_type = ensembl_content_type
        
        # Override content type if it is defined by function
        if "content_type" in func:
            content_type = func["content_type"]
        
        # Ovveride content type if it is provied when calling function
        if "content_type" in kwargs:
            content_type = kwargs["content_type"]
            del(kwargs["content_type"])
        
        #check the request type (GET or POST?)
        if func['method'] == 'GET':
            logger.debug("Submitting a GET request: url = '%s', headers = %s, params = %s"
                         % (url, {"Content-Type": content_type}, kwargs))
            
            # record this request
            self.last_url = url
            self.last_headers = {"Content-Type": content_type}
            self.last_params = kwargs
            self.last_data = {}
            self.last_method = "GET"
            self.last_attempt = 0
        
            resp = self.__get_response()
            
        elif func['method'] == 'POST':
            # in a POST request, separate post parameters from other parameters
            data = {}
            
            # pass key=value in POST data from kwargs
            for key in func['post_parameters']:
                if key in kwargs:
                    data[key] = kwargs[key]
                    del(kwargs[key])
                
            logger.debug("Submitting a POST request: url = '%s', headers = %s, params = %s, data = %s"
                         % (url, {"Content-Type": content_type}, kwargs, data))
            
            # record this request
            self.last_url = url
            self.last_headers = {"Content-Type": content_type}
            self.last_params = kwargs
            self.last_data = data
            self.last_method = "POST"
            self.last_attempt = 0
            
            resp = self.__get_response()
                
        else:
            raise NotImplementedError("Method '%s' not yet implemented" % (func['method']))
            
        #call response and return content
        return self.parseResponse(resp, content_type)
        
    # A function to get reponse from ensembl REST api
    def __get_response(self):
        """Call session get and post method. Return response"""
        
        # updating last_req time
        self.last_req = time.time()
        
        #Increment the request counter to rate limit requests    
        self.req_count += 1
        
        # Evaluating the numer of request in a second (according to EnsEMBL rest specification)
        if self.req_count >= self.reqs_per_sec:
            delta = time.time() - self.last_req

            # sleep upto wall_time
            if delta < self.wall_time:
                to_sleep = self.wall_time - delta
                logger.debug("waiting %s" % to_sleep)
                time.sleep(to_sleep)
                
            self.req_count = 0
        
        # my response
        resp = None
        
        # deal with exceptions
        try:
            # another request using the correct method
            if self.last_method == "GET":
                resp = self.session.get(
                    self.last_url,
                    headers=self.last_headers,
                    params=self.last_params,
                    timeout=self.timeout
                )
            elif self.last_method == "POST":
                # post parameters are load as POST data, other parameters are url parameters as GET requests
                resp = self.session.post(
                    self.last_url,
                    headers=self.last_headers,
                    data=json.dumps(self.last_data),
                    params=self.last_params,
                    timeout=self.timeout
                )
            # other methods are verifiedby others functions
                    
        except requests.ConnectionError as e:
            raise EnsemblRestServiceUnavailable(e)
                    
        except requests.Timeout as e:
            logger.error("%s request timeout: %s" % (self.last_method, e))
            
            # create a fake response in order to redo the query
            resp = namedtuple("fakeResponse", ["headers", "status_code", "text"])
            
            # add some data
            resp.headers = {}
            resp.status_code = 400
            resp.text = json.dumps({'message': repr(e), 'error': "%s timeout" % ensembl_user_agent})
        
        # return response
        return resp

    # A function to deal with a generic response
    def parseResponse(self, resp, content_type="application/json"):
        """Deal with a generic REST response"""
        
        logger.debug("Got %s" % resp.text)
        
        #record response for debug intent
        self.last_response = resp
        
        # initialize some values. Check if I'm rate limited
        self.rate_reset, self.rate_limit, self.rate_remaining, self.retry_after = self.__get_rate_limit(resp.headers)
        
        # parse status code
        if self.__check_retry(resp):
            return self.__retry_request()

        #handle content in different way relying on content-type
        if content_type == 'application/json':
            content = json.loads(resp.text)
        
        else:
            #default 
            content = resp.text
            
        return content
    
    def __check_retry(self, resp):
        """Parse status code and print warnings. Return True if a retry is needed"""
        
        # default status code
        message = ensembl_http_status_codes[resp.status_code][1]
        
        # parse status codes
        if resp.status_code > 304:
            ExceptionType = EnsemblRestError
            
            #Try to derive a more useful message than ensembl default message
            if resp.status_code == 400:
                json_message = json.loads(resp.text)
                if "error" in json_message:
                    message = json_message["error"]
                    
                #TODO: deal with special cases errors
                if message in ensembl_known_errors:
                    # call a function that will re-execute the REST request and then call again parseResponse
                    # if everithing is ok, a processed content is returned
                    logger.warning("EnsEMBL REST Service returned: %s" % message)
                    
                    # return true if retry needed
                    return True
            
            if resp.status_code == 429:
                ExceptionType = EnsemblRestRateLimitError

            raise ExceptionType(
                message,
                error_code=resp.status_code,
                rate_reset=self.rate_reset,
                rate_limit=self.rate_limit,
                rate_remaining=self.rate_remaining,
                retry_after=self.retry_after
            )
        
        # return a flag if status is ok
        return False

    @staticmethod
    def __get_rate_limit(headers):
        """Read rate limited attributes"""
        
        # initialize some values
        retry_after = None
        rate_reset = None
        rate_limit = None
        rate_remaining = None
        
        # for semplicity
        keys = [key.lower() for key in headers.keys()]
        
        if "X-RateLimit-Reset".lower() in keys:
            rate_reset = int(headers["X-RateLimit-Reset"])
            logger.debug("X-RateLimit-Reset: %s" % rate_reset)
            
        if "X-RateLimit-Limit".lower() in keys:
            rate_limit = int(headers["X-RateLimit-Limit"])
            logger.debug("X-RateLimit-Limit: %s" % rate_limit)
            
        if "X-RateLimit-Remaining".lower() in keys:
            rate_remaining = int(headers["X-RateLimit-Remaining"])
            logger.debug("X-RateLimit-Remaining: %s" % rate_remaining)
            
        if "Retry-After".lower() in keys:
            retry_after = float(headers["Retry-After"])
            logger.debug("Retry-After: %s" % retry_after)
            
        return rate_reset, rate_limit, rate_remaining, retry_after
        
    def __retry_request(self):
        """Retry last request in case of failure"""
        
        # update last attempt
        self.last_attempt += 1
        
        # a max of three attempts
        if self.last_attempt > self.max_attempts:
            # default status code
            message = ensembl_http_status_codes[self.last_response.status_code][1]
            
            # parse error if possible
            json_message = json.loads(self.last_response.text)
            if "error" in json_message:
                message = json_message["error"]
        
            raise EnsemblRestError("Max number of retries attempts reached. Last message was: %s"
                                   % message,
                                   error_code=self.last_response.status_code,
                                   rate_reset=self.rate_reset,
                                   rate_limit=self.rate_limit,
                                   rate_remaining=self.rate_remaining,
                                   retry_after=self.retry_after
                                   )
            
        # sleep a while. Increment on each attempt
        to_sleep = (self.wall_time + 1) * self.last_attempt
        
        logger.debug("Sleeping %s" % to_sleep)
        time.sleep(to_sleep)
    
        # another request using the correct method
        if self.last_method == "GET":
            #debug
            logger.debug("Retring last GET request (%s/%s): url = '%s', headers = %s, params = %s"
                         % (self.last_attempt, self.max_attempts, self.last_url, self.last_headers, self.last_params))
            
            resp = self.__get_response()
                
        elif self.last_method == "POST":
            #debug
            logger.debug("Retring last POST request (%s/%s): url = '%s', headers = %s, params = %s, data = %s" % (
                self.last_attempt,
                self.max_attempts,
                self.last_url,
                self.last_headers,
                self.last_params,
                self.last_data
            ))
            
            resp = self.__get_response()
        else:
            resp = None
        
        #call response and return content
        return self.parseResponse(resp, self.last_headers["Content-Type"])


class EnsemblGenomeRest(EnsemblRest):
    """EnsEMBL Genome REST API object"""

    def __init__(self, api_table=ensemblgenomes_api_table, base_url=ensembl_genomes_url, **kwargs):
        # Override default base_url
        kwargs["base_url"] = base_url
        
        # Call the Base Class init method
        EnsemblRest.__init__(self, api_table=api_table, **kwargs)



"""

    This file is part of pyEnsemblRest.
    Copyright (C) 2013-2016, Steve Moss

    pyEnsemblRest is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    pyEnsemblRest is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with pyEnsemblRest.  If not, see <http://www.gnu.org/licenses/>.

    EnsemblRest is a library for Python that wrap the EnsEMBL REST API.
    It simplifies all the API endpoints by abstracting them away from the
    end user and thus also ensures that an amendments to the library and/or
    EnsEMBL REST API won't cause major problems to the end user.

    Any questions, comments or issues can be addressed to gawbul@gmail.com.
"""

__author__ = "Steve Moss"
__copyright__ = "Copyright (C) 2013-2016, Steve Moss"
__credits__ = ["Steve Moss"]
__license__ = "GNU GPLv3"
__version__ = "0.2.3"
__maintainer__ = "Steve Moss"
__email__ = "gawbul@gmail.com"
__status__ = "beta"

from .ensemblrest import EnsemblRest, EnsemblGenomeRest
from .exceptions import EnsemblRestError, EnsemblRestRateLimitError, EnsemblRestServiceUnavailable




"""

    This file is part of pyEnsemblRest.
    Copyright (C) 2013-2016, Steve Moss

    pyEnsemblRest is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    pyEnsemblRest is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with pyEnsemblRest.  If not, see <http://www.gnu.org/licenses/>.

"""



from setuptools import setup
from setuptools import find_packages

__author__ = 'Steve Moss'
__email__ = 'gawbul@gmail.com'
__version__ = '0.2.3'

setup(
    # Basic package information.
    name='pyensemblrest',
    version=__version__,
    packages=find_packages(exclude=['test']),

    # Packaging options.
    include_package_data=True,

    # Package dependencies.
    install_requires=['requests>=1.0.0'],

    # testing modules
    test_suite = "test",

    # Metadata for PyPI.
    author='Steve Moss',
    author_email='gawbul@gmail.com',
    license='GPLv3',
    url='http://github.com/pyopensci/pyensemblrest/tree/master',
    keywords='ensembl python rest api',
    description='An easy way to access EnsEMBL data with Python.',
    long_description=open('README.rst').read(),
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Internet',
    ]
)



# -*- coding: utf-8 -*-
"""
Created on Wed May 18 14:25:55 2016

@author: Paolo Cozzi <paolo.cozzi@ptp.it>
"""

"""
    
    This file is part of pyEnsemblRest.
    Copyright (C) 2013-2016, Steve Moss

    pyEnsemblRest is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    pyEnsemblRest is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with pyEnsemblRest.  If not, see <http://www.gnu.org/licenses/>.

    This script allows testing of all pyEnsemblRest functionality.

"""

# importing modules
import time

# import my module
import ensemblrest

# import my exceptions
from ensemblrest.exceptions import EnsemblRestError, EnsemblRestRateLimitError, EnsemblRestServiceUnavailable

# import other modules
import unittest

import six

#TODO: test maximum POST input size
class EnsemblRest(unittest.TestCase):
    """A class to test EnsemblRest methods"""
    
    def setUp(self):
        """Create a EnsemblRest object"""
        self.EnsEMBL = ensemblrest.EnsemblRest()
        
    def tearDown(self):
        """Sleep a while before doing next request"""
        time.sleep(0.2)
        
    def test_BadRequest(self):
        """Do an ensembl bad request"""
        
        six.assertRaisesRegex(self, EnsemblRestError, "EnsEMBL REST API returned a 400 (Bad Request)*", self.EnsEMBL.getArchiveById, id="meow")
        
    def test_BadUrl(self):
        """Do a Not found request"""
        
        # record old uri value
        old_uri = self.EnsEMBL.getArchiveById.__globals__["ensembl_api_table"]["getArchiveById"]["url"]
        
        # set a new uri. This change a global value
        self.EnsEMBL.getArchiveById.__globals__["ensembl_api_table"]["getArchiveById"]["url"] = '/archive/meow/{{id}}'
        
        # do a request
        try:
            six.assertRaisesRegex(self, EnsemblRestError, "EnsEMBL REST API returned a 404 (Not Found)*", self.EnsEMBL.getArchiveById, id="ENSG00000157764")
            
        except AssertionError as e:
            # fix the global value
            self.EnsEMBL.getArchiveById.__globals__["ensembl_api_table"]["getArchiveById"]["url"] = old_uri
            # then raise exception
            raise Exception(e)
            
        # fix the global value
        self.EnsEMBL.getArchiveById.__globals__["ensembl_api_table"]["getArchiveById"]["url"] = old_uri
        
    def test_getMsg(self):
        """Do a bad request and get message"""

        msg = None
        try:
            self.EnsEMBL.getArchiveById(id="miao")
        except EnsemblRestError as e:
            msg = e.msg
        
        self.assertRegexpMatches(msg, "EnsEMBL REST API returned a 400 (Bad Request)*")
        
    def test_rateLimit(self):
        """Simulating a rate limiting environment"""
        
        # get a request
        self.EnsEMBL.getArchiveById(id="ENSG00000157764")
        
        # retrieve last_reponse
        response = self.EnsEMBL.last_response
        
        # get headers
        headers = response.headers
        
        # simulating a rate limiting
        # https://github.com/Ensembl/ensembl-rest/wiki/Rate-Limits#a-maxed-out-rate-limit-response
        headers["Retry-After"] = '40.0'
        headers["X-RateLimit-Limit"] = '55000'
        headers["X-RateLimit-Reset"] = '40'
        headers["X-RateLimit-Period"] = '3600'
        headers["X-RateLimit-Remaining"] = '0'
        
        # set a different status code
        response.status_code = 429
        
        # now parse request. headers is a reference to response.headers
        six.assertRaisesRegex(self, EnsemblRestRateLimitError, "EnsEMBL REST API returned a 429 (Too Many Requests)*", self.EnsEMBL.parseResponse, response)
        
        # try to read exception message
        msg = None
        try:
            self.EnsEMBL.parseResponse(response)
        
        except EnsemblRestError as e:
            msg = e.msg
        
        self.assertRegexpMatches(msg, "EnsEMBL REST API returned a 429 (Too Many Requests)*")
        
    def test_RestUnavailable(self):
        """Querying a not available REST server"""
        
        # get an ensembl rest service (sopposing that we have no local REST service)
        EnsEMBL = ensemblrest.EnsemblRest(base_url='http://localhost:3000')
        
        # get a request (GET)
        self.assertRaises(EnsemblRestServiceUnavailable, EnsEMBL.getArchiveById, id="ENSG00000157764")
        self.assertRaises(EnsemblRestServiceUnavailable, EnsEMBL.getArchiveByMultipleIds, id=["ENSG00000157764", "ENSG00000248378"])
        
    def test_SomethingBad(self):
        """raise exception when n of attempts exceeds"""
        
        # get a request
        self.EnsEMBL.getArchiveById(id="ENSG00000157764")
        
        # retrieve last_reponse
        response = self.EnsEMBL.last_response
        
        # raise last_attempt number
        self.EnsEMBL.last_attempt = self.EnsEMBL.max_attempts
        
        # create a fake request.Response class
        class FakeResponse():
            def __init__(self, resp):
                self.headers = resp.headers
                self.status_code = 400
                self.text = """{"error":"something bad has happened"}"""
                
        #instantiate a fake response
        fakeResponse = FakeResponse(response)
        
        # verify exception
        six.assertRaisesRegex(self, EnsemblRestError, "Max number of retries attempts reached.*", self.EnsEMBL.parseResponse, fakeResponse)
        
    def test_RequestTimeout(self):
        """Deal with connections timeout"""
        
        # get a new ensemblrest object
        ensGenomeRest = ensemblrest.EnsemblGenomeRest()
        
        # Ovverride max_attempts
        ensGenomeRest.max_attempts = 1
        ensGenomeRest.timeout = 1
        
        # verify exception
        six.assertRaisesRegex(self, EnsemblRestError, "Max number of retries attempts reached.* timeout", ensGenomeRest.getGeneFamilyById, id="MF_01687", compara="bacteria")


if __name__ == "__main__":
    unittest.main()
